@@ -1,7 +1,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2011 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2020 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -25,6 +25,7 @@
 
 #include "lib_mgt.h"
 #include "iso_tree.h"
+#include "iso_img.h"
 
 /*
 
@@ -32,7 +33,6 @@
 #include "xorriso_private.h"
 #include "xorrisoburn.h"
 
-#include "iso_img.h"
 #include "iso_manip.h"
 #include "sort_cmp.h"
 
@@ -183,9 +183,13 @@ int Xorriso_rename_suffix(struct XorrisO *xorriso, IsoNode *node, char *suffix,
 {
  int ret, lo= 0, ls= 0, strip_suffix;
  char *old_name= NULL, *show_name;
+ IsoImage *volume;
 
  strip_suffix= !!(flag & 2);
  
+ ret= Xorriso_get_volume(xorriso, &volume, 0);
+ if(ret <= 0)
+   goto ex;
  old_name= strdup((char *) iso_node_get_name(node));
  show_name= old_name;
  if(show_path != NULL)
@@ -205,7 +209,7 @@ int Xorriso_rename_suffix(struct XorrisO *xorriso, IsoNode *node, char *suffix,
      goto cannot_remove_suffix;
    strcpy(new_name, old_name);
    new_name[lo - ls]= 0;
-   ret = iso_node_set_name(node, new_name);
+   ret = iso_image_set_node_name(volume, node, new_name, 1);
    if (ret < 0) {
      Xorriso_process_msg_queues(xorriso,0);
      if (!(flag & 1))
@@ -235,7 +239,7 @@ cannot_append_suffix:;
      ret= 2 * (flag & 1); goto ex;
    } 
    sprintf(new_name, "%s%s", old_name, suffix);
-   ret = iso_node_set_name(node, new_name);
+   ret = iso_image_set_node_name(volume, node, new_name, 1);
    if (ret < 0) {
      Xorriso_process_msg_queues(xorriso,0);
      if (!(flag & 1))
@@ -270,7 +274,7 @@ int Xorriso_set_filter(struct XorrisO *xorriso, void *in_node,
  IsoExternalFilterCommand *cmd = NULL;
  char *old_name= NULL, *new_name= NULL, *suffix= "";
  IsoStream *stream;
- int is_renamed= 0;
+ IsoImage *volume;
 
  Xorriso_alloc_meM(new_name, char, SfileadrL);
  new_name[0]= 0;
@@ -337,7 +341,6 @@ int Xorriso_set_filter(struct XorrisO *xorriso, void *in_node,
                                (flag & 1) | (strip_suffix ? 2 : 0));
    if(ret <= 0 || ret == 2)
      goto ex;
-   is_renamed= 1;
  }
 
  if(strip_filter) {
@@ -375,6 +378,25 @@ int Xorriso_set_filter(struct XorrisO *xorriso, void *in_node,
    filter_ret= 1;
  } else if (internal_filter == 1 || internal_filter == 2) {
    filter_ret = iso_file_add_zisofs_filter(file, 1 | (internal_filter & 2));
+   if(filter_ret == (int) ISO_ZISOFS_TOO_MANY_PTR) {
+     /* Remove all buffered currently unused block pointers and try again */
+     ret= Xorriso_get_volume(xorriso, &volume, 0);
+     if(ret < 0)
+       goto ex;
+     if(ret > 0) {
+       Xorriso_msgs_submit(xorriso, 0,
+                           "Overflow of zisofs block pointers happened.",
+                           0, "WARNING", 0);
+       Xorriso_msgs_submit(xorriso, 0,
+                  "zisofs falling back to mode which needs 3 input read runs.",
+                           0, "WARNING", 0);
+       ret = iso_image_zisofs_discard_bpt(volume, 0);
+       if(ret <= 0)
+         goto ex;
+       filter_ret = iso_file_add_zisofs_filter(file,
+                                               1 | (internal_filter & 2));
+     }
+   }
    if(filter_ret < 0) {
      Xorriso_process_msg_queues(xorriso,0);
      if(!(internal_filter == 2 && filter_ret == (int) ISO_ZISOFS_WRONG_INPUT))
@@ -411,7 +433,10 @@ int Xorriso_set_filter(struct XorrisO *xorriso, void *in_node,
    }
  }
  if(filter_ret != 1 && new_name[0] && old_name != NULL) {
-   ret = iso_node_set_name(node, old_name);
+   ret= Xorriso_get_volume(xorriso, &volume, 0);
+   if(ret<=0)
+     goto ex;
+   ret = iso_image_set_node_name(volume, node, old_name, 1);
    if (ret < 0) {
      Xorriso_process_msg_queues(xorriso,0);
      if (!(flag & 1))
@@ -438,22 +463,15 @@ ex:;
 }
 
 
-/* @param flag bit0= delete filter with the given name
-*/
-int Xorriso_external_filter(struct XorrisO *xorriso,
-                            char *name, char *options, char *path, 
-                            int argc, char **argv, int flag)
+int Xorriso_external_filter_banned(struct XorrisO *xorriso, char *purpose,
+                                   int flag)
 {
- int ret, delete= 0, behavior= 0, extf_flag= 0, is_banned= 0;
- char *what, *what_next, *suffix= "";
- struct Xorriso_lsT *lst;
- struct Xorriso_extF *found_filter, *new_filter= NULL;
+ int is_banned= 0;
 
 #ifndef Xorriso_allow_external_filterS
  /* To be controlled by: configure --enable-external-filters */
 
- sprintf(xorriso->info_text, "%s : Banned at compile time.",
-         flag & 1 ? "-unregister_filter" : "-external_filter");
+ sprintf(xorriso->info_text, "%s : Banned at compile time.", purpose);
  Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
  sprintf(xorriso->info_text,
 "This may be changed at compile time by ./configure option --enable-external-filters");
@@ -476,16 +494,31 @@ int Xorriso_external_filter(struct XorrisO *xorriso,
  }
 #endif /* ! Xorriso_allow_extf_suiD */
 
- if(is_banned)
-   return(0);
-
  if(xorriso->filter_list_closed) {
    sprintf(xorriso->info_text,
-           "%s : Banned by previous command -close_filter_list",
-           flag & 1 ? "-unregister_filter" : "-external_filter");
+           "%s : Banned by previous command -close_filter_list", purpose);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
-   return(0);
+   is_banned= 1;
  } 
+ return(is_banned);
+}
+
+
+/* @param flag bit0= delete filter with the given name
+*/
+int Xorriso_external_filter(struct XorrisO *xorriso,
+                            char *name, char *options, char *path, 
+                            int argc, char **argv, int flag)
+{
+ int ret, delete= 0, behavior= 0, extf_flag= 0, is_banned= 0;
+ char *what, *what_next, *suffix= "";
+ struct Xorriso_lsT *lst;
+ struct Xorriso_extF *found_filter, *new_filter= NULL;
+
+ is_banned= Xorriso_external_filter_banned( xorriso,
+                      flag & 1 ? "-unregister_filter" : "-external_filter", 0);
+ if(is_banned)
+   return(0);
  if((!(flag & 1)) && path[0] != '/') {
    sprintf(xorriso->info_text,
            "-external_filter : Given command path does not begin by '/' : ");
@@ -642,13 +675,30 @@ int Xorriso_status_extf(struct XorrisO *xorriso, char *filter, FILE *fp,
  return(1);
 }
 
-
+/* @param flag bit0= iso_zisofs_set_params
+               bit1= iso_zisofs_ctrl_susp_z2
+*/
 int Xorriso_set_zisofs_params(struct XorrisO *xorriso, int flag)
 {
- int ret;
+ int ret, i;
  struct iso_zisofs_ctrl ctrl;
 
- ctrl.version= 0;
+ if(flag & 2) {
+   iso_zisofs_ctrl_susp_z2(!!xorriso->zisofs_susp_z2);
+ }
+ if(!(flag & 1))
+   return(1);
+
+ memset(&ctrl, 0, sizeof(ctrl));
+ ctrl.version= 1;
+ ret= iso_zisofs_get_params(&ctrl, 0);
+ if(ret < 0) {
+   Xorriso_report_iso_error(xorriso, "", ret,
+                  "Error when inquiring zisofs parameters before setting some",
+                  0, "FAILURE", 1);
+   return(0);
+ }
+
  ctrl.compression_level= xorriso->zlib_level;
  if(xorriso->zisofs_block_size == (1 << 16))
    ctrl.block_size_log2= 16;
@@ -656,6 +706,17 @@ int Xorriso_set_zisofs_params(struct XorrisO *xorriso, int flag)
    ctrl.block_size_log2= 17;
  else
    ctrl.block_size_log2= 15;
+
+ ctrl.v2_enabled= xorriso->zisofs_v2_enabled;
+ ctrl.max_total_blocks= xorriso->zisofs_max_total_blocks;
+ ctrl.max_file_blocks= xorriso->zisofs_max_file_blocks;
+ ctrl.v2_block_size_log2= 17;
+ for(i= 15; i <= 20; i++)
+   if(xorriso->zisofs_v2_block_size == (1 << i))
+     ctrl.v2_block_size_log2= i;
+ ctrl.block_number_target= xorriso->zisofs_block_number_target;
+ ctrl.bpt_discard_free_ratio= xorriso->zisofs_bpt_discard_free_ratio;
+
  ret= iso_zisofs_set_params(&ctrl, 0);
  Xorriso_process_msg_queues(xorriso,0);
  if(ret < 0) {
@@ -667,37 +728,96 @@ int Xorriso_set_zisofs_params(struct XorrisO *xorriso, int flag)
 }
 
 
+uint64_t Xorriso_zisofs_current_blocks(struct XorrisO *xorriso, int flag)
+{
+ int ret;
+ struct iso_zisofs_ctrl ctrl;
+
+ memset(&ctrl, 0, sizeof(ctrl));
+ ctrl.version= 1;
+ ret = iso_zisofs_get_params(&ctrl, 0);
+ if(ret != 1)
+   return(0);
+ return(ctrl.current_total_blocks);
+}
+
+
 int Xorriso_status_zisofs(struct XorrisO *xorriso, char *filter, FILE *fp,
                           int flag)
 /*
- bit0= do only report non-default settings
- bit1= do only report to fp
+ bit0= only report non-default settings
+ bit1= only report to fp
 */ 
 {
  off_t ziso_count= 0, osiz_count= 0;
  off_t gzip_count= 0, gunzip_count= 0;
+ uint64_t used_blocks;
+ int always= 0;
 
  iso_zisofs_get_refcounts(&ziso_count, &osiz_count, 0);
  iso_gzip_get_refcounts(&gzip_count, &gunzip_count, 0);
- if((flag & 1) && xorriso->zlib_level == xorriso->zlib_level_default &&
-    xorriso->zisofs_block_size == xorriso->zisofs_block_size_default &&
-    xorriso->zisofs_by_magic == 0 &&
-    ziso_count == 0 && osiz_count == 0 &&
-    gzip_count == 0 && gunzip_count == 0) {
-   if(filter == NULL)
-     return(2);
-   if(filter[0] == 0)
-     return 2;
+ if(!(flag & 1)) {
+   always= 1;
+ } else if(filter != NULL) {
+   if(filter[0] != 0)
+     always= 1;
  }
- sprintf(xorriso->result_line,
-     "-zisofs level=%d:block_size=%dk:by_magic=%s:ziso_used=%.f:osiz_used=%.f",
-     xorriso->zlib_level, xorriso->zisofs_block_size / 1024,
-     xorriso->zisofs_by_magic ? "on" : "off",
-     (double) ziso_count, (double) osiz_count);
- sprintf(xorriso->result_line + strlen(xorriso->result_line),
-         ":gzip_used=%.f:gunzip_used=%.f\n",
+
+ if(always || !(
+      xorriso->zlib_level == xorriso->zlib_level_default &&
+      xorriso->zisofs_block_size == xorriso->zisofs_block_size_default &&
+      xorriso->zisofs_by_magic == 0)) {
+   sprintf(xorriso->result_line,
+           "-zisofs level=%d:block_size=%dk:by_magic=%s\n",
+           xorriso->zlib_level, xorriso->zisofs_block_size / 1024,
+           xorriso->zisofs_by_magic == 1 ? "on" :
+           xorriso->zisofs_by_magic == 2 ? "v2" : "off");
+   Xorriso_status_result(xorriso, filter, fp, flag & 2);
+ }
+ if(always || !(
+      xorriso->zisofs_v2_enabled == 0 &&
+      xorriso->zisofs_v2_block_size == xorriso->zisofs_v2_block_size_default
+      && xorriso->zisofs_block_number_target == -1)){
+   sprintf(xorriso->result_line,
+          "-zisofs version_2=%s:block_size_v2=%dk:susp_z2=%s:bpt_target=%.f\n",
+           xorriso->zisofs_v2_enabled ? xorriso->zisofs_v2_enabled == 1 ?
+                                              "as_needed" : "on" : "off",
+           xorriso->zisofs_v2_block_size / 1024,
+           xorriso->zisofs_susp_z2 ? "on" : "off",
+           (double) xorriso->zisofs_block_number_target);
+   Xorriso_status_result(xorriso, filter, fp, flag & 2);
+ }
+ used_blocks= Xorriso_zisofs_current_blocks(xorriso, 0);
+ if(always || !(
+   xorriso->zisofs_max_total_blocks ==
+                                   xorriso->zisofs_max_total_blocks_default &&
+   xorriso->zisofs_max_file_blocks ==
+                                   xorriso->zisofs_max_file_blocks_default &&
+   xorriso->zisofs_bpt_discard_free_ratio ==
+                             xorriso->zisofs_bpt_discard_free_ratio_default)) {
+   sprintf(xorriso->result_line,
+          "-zisofs max_bpt=%.f:max_bpt_f=%.f:bpt_free_ratio=%.3f\n",
+          (double) xorriso->zisofs_max_total_blocks * 8.0,
+          (double) xorriso->zisofs_max_file_blocks * 8.0,
+          xorriso->zisofs_bpt_discard_free_ratio);
+   Xorriso_status_result(xorriso, filter, fp, flag & 2);
+ }
+ if(always || !(used_blocks == 0)) {
+   sprintf(xorriso->result_line,
+          "-zisofs bpt_used=%.f:bpt_ratio_used=%.3f\n",
+           (double) used_blocks * 8.0,
+           ((double) used_blocks) / (double) xorriso->zisofs_max_total_blocks);
+   Xorriso_status_result(xorriso, filter, fp, flag & 2);
+ }
+ if(always || !(
+    ziso_count == 0 && osiz_count == 0 &&
+    gzip_count == 0 && gunzip_count == 0)) {
+   sprintf(xorriso->result_line,
+         "-zisofs ziso_used=%.f:osiz_used=%.f:gzip_used=%.f:gunzip_used=%.f\n",
+         (double) ziso_count, (double) osiz_count,
          (double) gzip_count, (double) gunzip_count);
- Xorriso_status_result(xorriso, filter, fp, flag&2);
+   Xorriso_status_result(xorriso, filter, fp, flag & 2);
+ }
  return(1);
 }
 
